#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <random>

#define CL_HPP_TARGET_OPENCL_VERSION 120
#define CL_HPP_MINIMUM_OPENCL_VERSION 120
#include <CL/opencl.hpp>
#include <openssl/sha.h>

void fillstr(void *dst, std::string str) {
	const unsigned int len = str.length();
	const char *cstr = str.c_str();
	memcpy(dst,cstr,len);
}

int main(){
    //get all platforms (drivers)
    std::vector<cl::Platform> all_platforms;
    cl::Platform::get(&all_platforms);
    if(all_platforms.size()==0){
        std::cout<<" No platforms found. Check OpenCL installation!\n";
        exit(1);
    }
    cl::Platform default_platform=all_platforms[0];
    std::cout << "Using platform: "<<default_platform.getInfo<CL_PLATFORM_NAME>()<<"\n";
 
    //get default device of the default platform
    std::vector<cl::Device> all_devices;
    default_platform.getDevices(CL_DEVICE_TYPE_ALL, &all_devices);
    if(all_devices.size()==0){
        std::cout<<" No devices found. Check OpenCL installation!\n";
        exit(1);
    }
    cl::Device default_device=all_devices[0];
    std::cout<< "Using device: "<<default_device.getInfo<CL_DEVICE_NAME>()<<"\n";
 
 
    cl::Context context({default_device});
 
    cl::Program::Sources sources;
 
    // kernel calculates for each element C=A+B
	std::ifstream kernel_file("sha256.cl");
	std::stringstream kernel_code_buf;
	kernel_code_buf << kernel_file.rdbuf();
    std::string kernel_code = kernel_code_buf.str();
    sources.push_back({kernel_code.c_str(),kernel_code.length()});
 
    cl::Program program(context,sources);
    if(program.build({default_device})!=CL_SUCCESS){
        std::cout<<" Error building: "<<program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(default_device)<<"\n";
        exit(1);
    }
 
 
	unsigned char message[] = "Lowest Priviledge";
	unsigned int mlen = sizeof(message);
	unsigned char digest[32] = {};
    // create buffers on the device
	cl::Buffer message_buf(context,CL_MEM_READ_ONLY|CL_MEM_HOST_WRITE_ONLY,mlen);
	cl::Buffer mlen_buf(context,CL_MEM_READ_ONLY|CL_MEM_HOST_WRITE_ONLY,sizeof(unsigned int *));
	cl::Buffer digest_buf(context,CL_MEM_WRITE_ONLY|CL_MEM_HOST_READ_ONLY,sizeof(digest));
 
    //create queue to which we will push commands for the device.
    cl::CommandQueue queue(context,default_device);
 
    //write arrays A and B to the device
	queue.enqueueWriteBuffer(message_buf,CL_TRUE,0,mlen,message);
	queue.enqueueWriteBuffer(mlen_buf,CL_TRUE,0,sizeof(unsigned int *),&mlen);
 
    //run the kernel
    cl::Kernel kernel_sha256=cl::Kernel(program,"sha256");
    kernel_sha256.setArg(0,message_buf);
    kernel_sha256.setArg(1,mlen_buf);
    kernel_sha256.setArg(2,digest_buf);
    queue.enqueueNDRangeKernel(kernel_sha256,cl::NullRange,cl::NDRange(1),cl::NullRange);
    queue.finish();
 
	queue.enqueueReadBuffer(digest_buf,CL_TRUE,0,sizeof(digest),digest);

	std::cout<<"Result: shor string"<<std::endl;
	std::cout<<" message: \t"<<message<<std::endl<<" enhexed: \t";
	for(int i=0; i<mlen; i++) {
		const unsigned int c = static_cast<unsigned int>(message[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
	unsigned char reference[32];
	SHA256(message,mlen,reference);
    std::cout<<std::endl<<" expected: \t";
	for(int i=0; i<32; i++) {
		const unsigned int c = static_cast<unsigned int>(reference[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
    std::cout<<std::endl<<" actual: \t";
	for(int i=0; i<32; i++) {
		const unsigned int c = static_cast<unsigned int>(digest[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
	std::cout<<std::endl;

	std::random_device rand;
	unsigned int bytes_len = 2048;
	unsigned char bytes[bytes_len];
	for(unsigned int idx=0; idx<bytes_len; idx++) 
		bytes[idx] = static_cast<unsigned char>(0xff&rand());
	cl::Buffer bytes_buf(context,CL_MEM_READ_ONLY|CL_MEM_HOST_WRITE_ONLY,bytes_len);
	cl::Buffer blen_buf(context,CL_MEM_READ_ONLY|CL_MEM_HOST_WRITE_ONLY,sizeof(unsigned int *));
	queue.enqueueWriteBuffer(bytes_buf,CL_TRUE,0,bytes_len,bytes);
	queue.enqueueWriteBuffer(blen_buf,CL_TRUE,0,sizeof(unsigned int *),&bytes_len);
    kernel_sha256.setArg(0,bytes_buf);
    kernel_sha256.setArg(1,blen_buf);
    kernel_sha256.setArg(2,digest_buf);
    queue.enqueueNDRangeKernel(kernel_sha256,cl::NullRange,cl::NDRange(1),cl::NullRange);
    queue.finish();
 
	queue.enqueueReadBuffer(digest_buf,CL_TRUE,0,sizeof(digest),digest);

	std::cout<<"Result: long random bytes"<<std::endl;
	std::cout<<" enhexed: \t";
	for(unsigned int i=0; i<std::min(29U,bytes_len); i++) {
		const unsigned int c = static_cast<unsigned int>(bytes[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
	if(bytes_len>32) std::cout<<"...";
	SHA256(bytes,bytes_len,reference);
    std::cout<<std::endl<<" expected: \t";
	for(int i=0; i<32; i++) {
		const unsigned int c = static_cast<unsigned int>(reference[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
    std::cout<<std::endl<<" actual: \t";
	for(int i=0; i<32; i++) {
		const unsigned int c = static_cast<unsigned int>(digest[i]);
		std::cout<<std::hex<<std::setfill('0')<<std::setw(2)<<c;
	}
	std::cout<<std::endl;

    return 0;
}
